﻿using System.Text.RegularExpressions;

namespace ResticLogParser.src
{
    internal class LogParser
    {
        // Example match:
        // new       /E/Projects/Unity/SpriteTransfer/Images/Equipment/Belts/Belt 3 Up.png, saved in 0.000s (7.201 KiB added)
        private Regex _newFileLogEntryRegex = new(@"new\s+(.+), saved in \d+\.\d+s \((\d+(?:\.\d+)? .+) added\)");

        // Example match:
        // modified  /C/Users/warrence/AppData/Local/Log/Desktop.log, saved in 0.009s (1.135 MiB added, 1.135 MiB stored)
        private Regex _modifiedFileLogEntryRegex = 
            new(@"modified\s+(.+), saved in \d+\.\d+s \((\d+(?:\.\d+)? .+) added, (\d+(?:\.\d+)? .+) stored\)");

        private string _logFileName;
        private readonly List<FileLogEntry> _files = new();
        private readonly List<DirectoryNode> _directories = new();
        private readonly DirectoryNode _root = new(null, "/", "/");

        public LogParser(string logFileName)
        {
            if (string.IsNullOrEmpty(logFileName))
            {
                throw new ArgumentNullException(nameof(logFileName));
            }

            _logFileName = logFileName;
        }

        public void Run()
        {
            Console.WriteLine($"Running restic log parser for '{_logFileName}'...");
            ParseLogFile();
            WriteFileInformation();
            WriteDirectoryInformation();
        }

        private void ParseLogFile()
        {
            try
            {
                using StreamReader reader = new(_logFileName);
                while (!reader.EndOfStream)
                {
                    string line = reader.ReadLine()!;
                    TryParseLogEntry(line);
                }
            }
            catch (IOException e)
            {
                Console.WriteLine("The file could not be read:");
                Console.WriteLine(e.Message);
            }
        }

        private bool TryParseLogEntry(string logLine)
        {
            Match match = _newFileLogEntryRegex.Match(logLine);
            if (match.Success)
            {
                var size = FileSize.GetBytesFromString(match.Groups[2].Value);
                AddFileLogEntry(new FileLogEntry(match.Groups[1].Value, FileLogEntryType.New, size));
                return true;
            }

            match = _modifiedFileLogEntryRegex.Match(logLine);
            if (match.Success)
            {
                var addedSize = FileSize.GetBytesFromString(match.Groups[2].Value);
                var storedSize = FileSize.GetBytesFromString(match.Groups[3].Value);
                AddFileLogEntry(new FileLogEntry(match.Groups[1].Value, FileLogEntryType.Modified, addedSize, storedSize));
                return true;
            }

            return false;
        }

        private void AddFileLogEntry(FileLogEntry fileLogEntry)
        {
            _files.Add(fileLogEntry);

            string[] split = fileLogEntry.FileName.Split('/', StringSplitOptions.RemoveEmptyEntries);
            DirectoryNode parent = _root;
            // Skips the last one, since it is the file name.
            for (int i = 0; i < split.Length - 1; i++)
            {
                var child = parent.FindChild(split[i]);
                if (child == null)
                {
                    child = parent.AddChild(split[i], $"{parent.FullName}{split[i]}/");
                    _directories.Add(child);
                }
                child.AddedSize += fileLogEntry.AddedSize;
                parent = child;
            }
            parent.FileCount++;
        }

        private void WriteFileInformation()
        {
            string outputFileName = Path.Combine(Path.GetDirectoryName(_logFileName)!, "restic-added-data-files.txt");
            Console.WriteLine($"Writing added files to '{outputFileName}'...");

            _files.Sort();

            using (StreamWriter outputFile = new StreamWriter(outputFileName))
            {
                for (int i = _files.Count - 1; i >= 0; i--)
                {
                    outputFile.WriteLine(_files[i].ToString());
                }
            }
        }

        private void WriteDirectoryInformation()
        {
            string outputFileName = Path.Combine(Path.GetDirectoryName(_logFileName)!, "restic-added-data-dirs.txt");
            Console.WriteLine($"Writing directory infos to '{outputFileName}'...");

            _directories.Sort();

            using (StreamWriter outputFile = new StreamWriter(outputFileName))
            {
                for (int i = _directories.Count - 1; i >= 0; i--)
                {
                    if (_directories[i].FileCount > 0)
                    {
                        outputFile.WriteLine(_directories[i].ToString());
                    }
                }
            }
        }
    }
}
